from __future__ import annotations

from typing import List

import dilib


def get_db_value(_0, _1) -> bool:
    return False


class Seat:
    pass


class Engine:
    @property
    def started(self) -> bool:
        raise NotImplementedError()

    def start(self):
        pass


class DBEngine(Engine, dilib.SingletonMixin):
    def __init__(self, db_address: str):
        self.db_address = db_address

    @property
    def started(self) -> bool:
        return get_db_value(self.db_address, "engine")


class MockEngine(Engine, dilib.SingletonMixin):
    @property
    def started(self) -> bool:
        return True


class Car(dilib.SingletonMixin):
    def __init__(self, seats: List[Seat], engine: Engine):
        self.seats = seats
        self.engine = engine

        self.state = 0

    def drive(self):
        if not self.engine.started:
            self.engine.start()
        self.state = 1

    def stop(self):
        self.state = 0


# noinspection PyTypeChecker
class EngineConfig(dilib.Config):

    db_address = dilib.GlobalInput(type_=str, default="ava-db")
    engine = DBEngine(db_address)  # type: ignore


# noinspection PyTypeChecker
class CarConfig(dilib.Config):

    engine_config = EngineConfig()

    seat_cls = dilib.Object(Seat)
    seats = dilib.Prototype(
        lambda cls, n: [cls() for _ in range(n)], seat_cls, 2
    )

    car = Car(seats, engine=engine_config.engine)  # type: ignore


def test_basic_demo():
    config = CarConfig().get(db_address="ava-db")
    container = dilib.Container(config)

    car = container.car
    assert isinstance(car, Car)
    assert id(car) == id(container.car)  # Because it's a Singleton
    assert isinstance(car.engine, DBEngine)


def test_perturb_demo():
    config = CarConfig().get(db_address="ava-db")
    config.engine_config.engine = MockEngine()
    container = dilib.Container(config)

    assert isinstance(container.car.engine, MockEngine)



from __future__ import annotations

from typing import Any, Dict, Iterable, Optional, Union, cast

from dilib import di_config, specs, utils


class ConfigProxy:
    """Read-only helper to marshal config values."""

    def __init__(self, container: Container, config: di_config.Config):
        self.container = container
        self.config = config

    def __getattr__(self, key: str) -> Any:
        # noinspection PyProtectedMember
        return self.container._get(self.config, key)

    def __getitem__(self, key: str) -> Any:
        return utils.nested_getattr(self, key)

    def __dir__(self) -> Iterable[str]:
        return dir(self.config)


class Container:
    """Getter of (cached) instances."""

    def __init__(self, config: di_config.Config):
        if isinstance(config, di_config.ConfigSpec):
            raise ValueError(
                "Expected Config type, got ConfigSpec. "
                "Please call .get() on the config."
            )

        self._config = config
        self._config.freeze()

        self._instance_cache: Dict[Union[str, int], Any] = {}

    def _process_arg(self, config: di_config.Config, arg: Any) -> Any:
        if isinstance(arg, specs.Spec):
            # noinspection PyProtectedMember
            if arg.spec_id in config._keys:
                # noinspection PyProtectedMember
                config_key = config._keys[arg.spec_id]
                result = self._get(config, config_key)
            elif isinstance(arg, specs.Prototype):
                # Anonymous prototype or singleton
                result = self._materialize_spec(config, arg).instantiate()
            else:
                raise ValueError(f"Unrecognized arg type: {type(arg)}")
        elif isinstance(arg, specs.AttrFuture):
            # noinspection PyProtectedMember
            config_key = config._keys[arg.parent_spec_id]
            result = self._get(config, config_key)

            for attr in arg.attrs:
                result = getattr(result, attr)
        elif isinstance(arg, (tuple, list)):
            result = type(arg)(self._process_arg(config, elem) for elem in arg)
        elif isinstance(arg, dict):
            result = {k: self._process_arg(config, v) for k, v in arg.items()}
        else:
            result = arg

        return result

    def _materialize_spec(
        self, config: di_config.Config, spec: specs.Prototype
    ) -> specs.Prototype:
        """Return Spec copy with materialized args/kwargs."""
        materialized_args = [
            self._process_arg(config, arg) for arg in spec.args
        ]
        materialized_kwargs = {
            key: self._process_arg(config, arg)
            for key, arg in spec.kwargs.items()
        }
        if spec.lazy_kwargs:
            materialized_lazy_kwargs = self._process_arg(
                config, spec.lazy_kwargs
            )
            materialized_kwargs.update(
                {
                    key: self._process_arg(config, arg)
                    for key, arg in materialized_lazy_kwargs.items()
                }
            )

        return spec.copy_with(*materialized_args, **materialized_kwargs)

    def _get(self, config: di_config.Config, key: str) -> Any:
        """Get instance represented by key in given config."""
        spec = getattr(config, key)
        if isinstance(spec, specs.Object):
            return spec.obj
        elif isinstance(spec, specs.Singleton):
            try:
                return self._instance_cache[spec.spec_id]
            except KeyError:
                pass

            instance = self._materialize_spec(config, spec).instantiate()
            self._instance_cache[spec.spec_id] = instance
            return instance
        elif isinstance(spec, specs.Prototype):
            spec = cast(specs.Prototype, spec)
            return self._materialize_spec(config, spec).instantiate()
        elif isinstance(spec, di_config.Config):
            return ConfigProxy(self, spec)
        elif isinstance(spec, specs.AttrFuture):
            # noinspection PyProtectedMember
            key = config._keys[spec.parent_spec_id]
            obj = self._get(config, key)

            for idx, attr in enumerate(spec.attrs):
                obj = getattr(obj, attr)
                if idx == len(spec.attrs) - 1:
                    return obj

            raise ValueError(
                f"Failed to resolve attr reference: "
                f"spec_id={spec.spec_id}, attrs={spec.attrs}"
            )
        else:
            raise ValueError(f"Unrecognized spec type: {type(spec)}")

    def get(self, key: str, default: Optional[Any] = None) -> Any:
        if key in dir(self):
            return self[key]

        return default

    def clear(self):
        """Clear instance cache."""
        self._instance_cache.clear()

    def __getattr__(self, key: str) -> Any:
        return self._get(self._config, key)

    def __getitem__(self, key: str) -> Any:
        return utils.nested_getattr(self, key)

    def __dir__(self) -> Iterable[str]:
        return dir(self._config)



# flake8: noqa
from dilib.container import ConfigProxy, Container
from dilib.di_config import Config
from dilib.errors import (
    ConfigError,
    FrozenConfigError,
    InputConfigError,
    NewKeyConfigError,
    SetChildConfigError,
)
from dilib.specs import (
    Forward,
    GlobalInput,
    LocalInput,
    Object,
    Prototype,
    PrototypeIdentity,
    PrototypeMixin,
    Singleton,
    SingletonDict,
    SingletonList,
    SingletonMixin,
    SingletonTuple,
    Spec,
    SpecID,
)



import pytest

import dilib
from dilib.tests import test_config


def test_basic():
    config = test_config.BasicConfig().get()
    container = dilib.Container(config)

    assert container.x == 1
    assert container.y == 2
    assert isinstance(container.foo, test_config.SingletonValue)
    assert container.foo.value == 1
    assert isinstance(container.bar, test_config.PrototypeValue)
    assert container.bar.value == 2

    assert container.foo is container.foo  # foo is a Singleton
    assert container.bar is not container.bar  # foo is a Prototype


def test_perturb_basic():
    config = test_config.BasicConfig().get()
    config.x = 2

    container = dilib.Container(config)
    assert container.y == 3
    assert container.foo.value == 2
    assert container.bar.value == 3


def test_get_nested():
    config = test_config.GrandParentConfig().get()
    container = dilib.Container(config)

    assert container.parent_config0.basic_config.x == 1
    assert container["parent_config0.basic_config.x"] == 1


def test_dir():
    config = test_config.GrandParentConfig().get()
    container = dilib.Container(config)

    assert dir(container) == [
        "foobar",
        "parent_config0",
        "parent_config1",
        "some_str0",
    ]
    assert dir(container.parent_config0) == ["basic_config", "baz0"]


def test_perturb_nested():
    config = test_config.GrandParentConfig().get()
    config.parent_config0.basic_config.x = 10

    container = dilib.Container(config)

    assert container.parent_config0.basic_config.x == 10
    assert container.parent_config0.basic_config.foo.value == 10
    assert container.parent_config0.basic_config.bar.value == 11
    assert (
        container.parent_config0.basic_config.foo
        is container.parent_config1.basic_config.foo
    )
    assert container.parent_config0.baz0.value == 10
    assert container.parent_config1.baz1.value == 10
    assert container.foobar.value == 10


def test_nested_keyerror():
    config = test_config.ErrorGrandParentConfig().get()
    container = dilib.Container(config)

    with pytest.raises(KeyError):
        try:
            container.foobar
        except KeyError as exc:
            assert str(exc) == (
                "\"<class 'dilib.tests.test_config.ParentConfig0'>: "
                "'non_existent_field'\""
            )
            raise


def test_input_config():
    config = test_config.InputConfig1().get(name="hi")
    config.input_config0.x = 100

    container = dilib.Container(config)
    assert container.input_config0.name == "hi"
    assert container.input_config0.context == "default"
    assert container.input_config0.x == 100
    assert container.y == 101


def test_collection_config():
    config = test_config.CollectionConfig().get()
    container = dilib.Container(config)

    assert container.x == 1
    assert container.y == 2
    assert container.foo_tuple == (1, 2)
    assert container.foo_list == [1, 2]
    assert container.foo_dict_kwargs == {"x": 1, "y": 2}
    assert container.foo_dict_values0 == {1: 1, 2: 2}
    assert container.foo_dict_values1 == {"values": 1}

    assert container.foo_tuple is container.foo_tuple


def test_anonymous():
    config = test_config.AnonymousConfig().get()
    container = dilib.Container(config)

    assert container.y.value.value is container.x
    assert container.z.value.value is container.x


def test_underscore():
    config = test_config.WrapperConfig().get()
    container = dilib.Container(config)

    assert container.value.value is container._value


def test_forward():
    config = test_config.ForwardConfig().get()
    container = dilib.Container(config)

    assert (
        container.foo is container.other_config.parent_config0.basic_config.foo
    )
    assert (
        container.foo_value.value
        is container.other_config.parent_config0.basic_config.foo
    )


def test_perturb_forward():
    config = test_config.ForwardConfig().get()

    config.x = 1000

    container = dilib.Container(config)

    # Objs that depend on original x remain unperturbed, but objs
    # that depend on the forward alias are perturbed.
    assert container.other_config.parent_config0.basic_config.x == 1
    assert container.other_config.foobar.value == 1
    assert container.x == 1000
    assert container.x_value.value == 1000


def test_perturb_partial_kwargs():
    config = test_config.PartialKwargsConfig().get()

    config.x = 10
    config.y = 20

    container = dilib.Container(config)

    assert container.x == 10
    assert container.y == 20
    assert container.values.x == 10
    assert container.values.y == 20
    assert container.values.z == 10


def test_perturb_partial_kwargs_other():
    config = test_config.PartialKwargsOtherConfig().get()

    config.partial_kwargs_config.x = 10

    container = dilib.Container(config)

    assert container.partial_kwargs_config.values.x == 10
    assert container.partial_kwargs_config.values.y == 2
    assert container.partial_kwargs_config.values.z == 10

    assert container.values.x == 10
    assert container.values.y == 2
    assert container.values.z == 3


class ObjWithAttr:
    @property
    def test_attr(self) -> int:
        return 1


class ObjAttrConfig(dilib.Config):
    test_obj = dilib.Singleton(ObjWithAttr)
    test_obj_attr = dilib.Forward(test_obj.test_attr)


class NestedConfigObjAttrConfig(dilib.Config):
    cfg = ObjAttrConfig()

    test_obj = dilib.Forward(cfg.test_obj)
    test_obj_attr = dilib.Forward(cfg.test_obj_attr)


def test_obj_attr():
    config = ObjAttrConfig().get()
    container = dilib.Container(config)

    assert container.test_obj.test_attr == 1
    assert container.test_obj_attr == 1


def test_nested_config_obj_attr():
    config = NestedConfigObjAttrConfig().get()
    container = dilib.Container(config)

    assert container.test_obj.test_attr == 1
    assert container.test_obj_attr == 1
    assert container.test_obj is container.cfg.test_obj



class ConfigError(RuntimeError):
    pass


class FrozenConfigError(ConfigError):
    pass


class InputConfigError(ConfigError):
    pass


class NewKeyConfigError(ConfigError):
    pass


class SetChildConfigError(ConfigError):
    pass



import types
from typing import Any

import pytest

import dilib


class Value:
    def __init__(self, value: Any):
        self.value = value


class Values:
    def __init__(self, x: Any, y: Any, z: Any):
        self.x = x
        self.y = y
        self.z = z


class SingletonValue(dilib.SingletonMixin):
    def __init__(self, value: int):
        self.value = value


class PrototypeValue(dilib.PrototypeMixin):
    def __init__(self, value: int):
        self.value = value


# noinspection PyTypeChecker
class BasicConfig(dilib.Config):

    x = dilib.Object(1)
    y = dilib.Prototype(lambda x, offset: x + offset, x, offset=1)

    foo = SingletonValue(value=x)  # type: ignore
    bar = PrototypeValue(value=y)  # type: ignore


def test_config_spec():
    # No inputs
    assert BasicConfig() == BasicConfig()
    assert hash(BasicConfig()) == hash(BasicConfig())

    # Basic inputs
    assert BasicConfig(x=1, y="hi") == BasicConfig(x=1, y="hi")
    assert BasicConfig(x=1, y="hi") != BasicConfig()

    assert hash(BasicConfig(x=1, y="hi")) == hash(BasicConfig(x=1, y="hi"))
    assert hash(BasicConfig(x=1, y="hi")) != hash(BasicConfig())


def test_basic():
    config = BasicConfig().get()

    assert config.x.obj == 1
    assert isinstance(config.y.cls, types.LambdaType)
    assert config.foo.cls is SingletonValue  # noqa
    assert config.bar.cls is PrototypeValue  # noqa


def test_perturb_basic():
    config0 = BasicConfig().get()

    config0.x = 2
    assert config0.x.obj == 2  # type: ignore

    # No class-level interactions
    config1 = BasicConfig().get()

    assert config1.x.obj == 1  # noqa


def test_perturb_after_freeze():
    config = BasicConfig().get()

    config.freeze()
    with pytest.raises(dilib.FrozenConfigError):
        config.x = 100


def test_add_key_after_load():
    config = BasicConfig().get()

    with pytest.raises(dilib.NewKeyConfigError):
        config.new_x = 100


# noinspection PyTypeChecker
class ParentConfig0(dilib.Config):

    basic_config = BasicConfig()

    baz0 = SingletonValue(basic_config.x)  # type: ignore


# noinspection PyTypeChecker
class ParentConfig1(dilib.Config):

    basic_config = BasicConfig()

    baz1 = SingletonValue(basic_config.x)  # type: ignore
    some_str1 = dilib.Object("abc")


# noinspection PyTypeChecker
class GrandParentConfig(dilib.Config):

    parent_config0 = ParentConfig0()
    parent_config1 = ParentConfig1()

    foobar = SingletonValue(parent_config0.basic_config.x)  # type: ignore
    some_str0 = dilib.Object("hi")


# noinspection PyTypeChecker
class ErrorGrandParentConfig(dilib.Config):

    parent_config0 = ParentConfig0()
    parent_config1 = ParentConfig1()

    # This is pointing to a non-existent attr, so we will fail when
    # trying to get foobar via the container.
    foobar = dilib.Forward(parent_config0.non_existent_field)


def test_dir():
    config = GrandParentConfig().get()

    assert dir(config) == [
        "foobar",
        "parent_config0",
        "parent_config1",
        "some_str0",
    ]
    assert dir(config.parent_config0) == ["basic_config", "baz0"]


def test_nested_config():
    config = GrandParentConfig().get()

    assert id(config.parent_config0.basic_config) == id(
        config.parent_config1.basic_config
    )


def test_perturb_nested_config_attrs():
    config = GrandParentConfig().get()

    config.some_str0 = "hello"
    config.parent_config0.basic_config.x = 100
    config.parent_config1.some_str1 = "def"

    # noinspection PyUnresolvedReferences
    assert config.some_str0.obj == "hello"  # type: ignore
    assert config.parent_config1.basic_config.x.obj == 100
    # noinspection PyUnresolvedReferences
    assert config.parent_config1.some_str1.obj == "def"  # type: ignore


def test_perturb_nested_config_strs():
    config = GrandParentConfig().get()

    config["some_str0"] = "hello"
    config["parent_config0.basic_config.x"] = 100
    config["parent_config1.some_str1"] = "def"

    assert config["some_str0"].obj == "hello"
    assert config["parent_config1.basic_config.x.obj"] == 100
    assert config["parent_config1.some_str1"].obj == "def"


def test_perturb_nested_child_config():
    config = GrandParentConfig().get()

    with pytest.raises(dilib.SetChildConfigError):
        config.parent_config0 = ParentConfig1()


class InputConfig0(dilib.Config):

    name = dilib.GlobalInput(str)
    context = dilib.GlobalInput(str, default="default")
    x = dilib.LocalInput(int)


class InputConfig1(dilib.Config):

    input_config0 = InputConfig0(x=1)

    y = dilib.Prototype(
        lambda x, offset: x + offset, input_config0.x, offset=1
    )


class BadInputConfig(dilib.Config):

    input_config0 = InputConfig0()  # Note missing inputs


def test_input_config():
    with pytest.raises(dilib.InputConfigError):
        InputConfig1().get()

    with pytest.raises(dilib.InputConfigError):
        InputConfig1().get(name=1)

    with pytest.raises(dilib.InputConfigError):
        BadInputConfig().get(name="hi")

    config = InputConfig1().get(name="hi")

    assert config.input_config0.name.obj == "hi"
    assert config.input_config0.context.obj == "default"
    assert config.input_config0.x.obj == 1


class CollectionConfig(dilib.Config):

    x = dilib.Object(1)
    y = dilib.Object(2)

    foo_tuple = dilib.SingletonTuple(x, y)
    foo_list = dilib.SingletonList(x, y)
    foo_dict_kwargs = dilib.SingletonDict(x=x, y=y)
    foo_dict_values0 = dilib.SingletonDict({1: x, 2: y})
    foo_dict_values1 = dilib.SingletonDict({"values": x})


class AnonymousConfig(dilib.Config):

    x = dilib.Singleton(Value, 1)
    y = dilib.Singleton(Value, dilib.Singleton(Value, x))
    z = dilib.Singleton(Value, dilib.Prototype(Value, x))


class WrapperConfig(dilib.Config):

    _value = dilib.Singleton(Value, 1)
    value = dilib.Singleton(Value, _value)


class ForwardConfig(dilib.Config):

    other_config = GrandParentConfig()

    x = dilib.Forward(other_config.parent_config0.basic_config.x)
    x_value = dilib.Singleton(Value, value=x)

    foo = dilib.Forward(other_config.parent_config0.basic_config.foo)
    foo_value = dilib.Singleton(Value, value=foo)


class PartialKwargsConfig(dilib.Config):

    x = dilib.Object(1)
    y = dilib.Object(2)

    partial_kwargs = dilib.SingletonDict(x=x, y=y)

    values = dilib.Singleton(Values, z=x, __lazy_kwargs=partial_kwargs)


class PartialKwargsOtherConfig(dilib.Config):

    partial_kwargs_config = PartialKwargsConfig()

    z = dilib.Object(3)
    values = dilib.Singleton(
        Values, z=z, __lazy_kwargs=partial_kwargs_config.partial_kwargs
    )


def test_extra_global_inputs():
    with pytest.raises(dilib.InputConfigError):
        try:
            InputConfig1().get(name="testing", foobar=123)
        except dilib.InputConfigError as exc:
            assert "extra" in str(exc) and "'foobar'" in str(exc)
            raise


class InputConfigWithCollision(dilib.Config):

    input_config0 = InputConfig0(x=1)

    # "name" collides with input_config0.name
    name = dilib.GlobalInput(str)


def test_global_input_collisions():
    with pytest.raises(dilib.InputConfigError):
        try:
            InputConfigWithCollision().get(name="testing")
        except dilib.InputConfigError as exc:
            assert "collision" in str(exc) and "'name'" in str(exc)
            raise



from typing import Any


def nested_getattr(obj: Any, key: str) -> Any:
    for key_part in key.split("."):
        obj = getattr(obj, key_part)
    return obj


def nested_setattr(obj: Any, key: str, value: Any):
    split_key = key.split(".")
    for index, key_part in enumerate(split_key):
        if index < len(split_key) - 1:
            obj = getattr(obj, key_part)
        else:
            setattr(obj, key_part, value)



from __future__ import annotations

import datetime
from typing import Any, Dict, Iterable, Optional, Set, Type, Union

from dilib import errors, specs, utils

PRIMITIVE_TYPES = (
    type(None),
    bool,
    int,
    float,
    str,
    datetime.date,
    datetime.time,
    datetime.datetime,
)


def check_type(
    value: Any, type_: Optional[Type] = None, tag: Optional[str] = None
):
    if type_ is None:
        return

    if hasattr(type_, "__args__"):
        # TODO! Check nested typing types here.
        return
    else:
        types = (type_,)

    if not isinstance(value, types):
        raise errors.InputConfigError(
            f"{tag} input mismatch types: {type(value)} is not {type_}"
        )


class ConfigSpec(specs.Spec):
    """Spec for Configs."""

    def __init__(self, cls: Type[Config], **kwargs):
        super().__init__()
        self.cls = cls
        self.kwargs = kwargs

    def _instantiate(self, config_locator: ConfigLocator) -> Config:
        """Instantiate, useful when using outside DI framework."""
        return specs.instantiate(self.cls, config_locator, **self.kwargs)

    def get(self, **global_inputs) -> Config:
        """Get instance of this Config."""
        config_locator = ConfigLocator(**global_inputs)
        config = config_locator.get(self)

        # noinspection PyProtectedMember
        global_input_keys = config._get_all_global_input_keys()
        extra_global_input_keys = set(global_inputs.keys()) - global_input_keys
        if extra_global_input_keys:
            raise errors.InputConfigError(
                f"Provided extra global inputs "
                f"not specified in configs: {extra_global_input_keys}"
            )

        return config

    def __eq__(self, other: Any) -> bool:
        return (
            type(other) is ConfigSpec
            and self.cls is other.cls
            and self.kwargs == other.kwargs
        )

    def __hash__(self) -> int:
        return hash(
            (
                self.__class__.__name__,
                frozenset(self.kwargs.items()),
            )
        )


class Config:
    """dataclass-style description of params and obj graph."""

    _INTERNAL_FIELDS = (
        "_config_locator",
        "_keys",
        "_specs",
        "_child_configs",
        "_global_inputs",
        "_loaded",
        "_frozen",
        "_get_all_global_input_keys",
        "_process_input",
        "_load",
        "freeze",
    )

    def __new__(cls: Type, **kwargs):
        # noinspection PyTypeChecker
        return ConfigSpec(cls, **kwargs)

    def __init__(
        self, config_locator: Optional[ConfigLocator] = None, **local_inputs
    ):
        if config_locator is None:
            raise ValueError(
                "config_locator is only defaulted to "
                "None to aid PyCharm inspections. "
                "Use config's cls.get() instead."
            )
        self._config_locator = config_locator

        for value in local_inputs.values():
            if not isinstance(value, PRIMITIVE_TYPES):
                raise errors.InputConfigError(
                    f"Unsupported local input type: {type(value)}"
                )

        self._keys: Dict[specs.SpecID, str] = {}
        self._specs: Dict[str, specs.Spec] = {}
        self._child_configs: Dict[str, Config] = {}
        self._global_inputs: Dict[str, specs.SpecID] = {}

        self._loaded = False
        self._frozen = False

        self._load(**local_inputs)

    # For mypy
    def __call__(self, *args, **kwargs):
        return None

    def _get_all_global_input_keys(
        self, all_global_input_keys: Optional[Dict[str, specs.SpecID]] = None
    ) -> Set[str]:
        """Recursively get all global input keys of this config and all child
        configs."""
        all_global_input_keys = (
            all_global_input_keys if all_global_input_keys is not None else {}
        )

        for key, spec_id in self._global_inputs.items():
            if key in all_global_input_keys:
                if all_global_input_keys[key] != spec_id:
                    raise errors.InputConfigError(
                        f"Found global input collision: {key!r}"
                    )

            all_global_input_keys[key] = spec_id

        for key, child_config in self._child_configs.items():
            # noinspection PyProtectedMember
            child_config._get_all_global_input_keys(all_global_input_keys)

        return set(all_global_input_keys.keys())

    def _process_input(
        self, key: str, spec: specs.Input, inputs: Dict[str, Any], tag: str
    ) -> specs.Object:
        """Convert Input spec to Object spec."""
        try:
            value = inputs[key]
        except KeyError:
            if spec.default != specs.MISSING:
                value = spec.default
            else:
                raise errors.InputConfigError(f"{tag} input not set: {key!r}")

        check_type(value, spec.type_, tag=tag)

        new_spec = specs.Object(value)
        new_spec.spec_id = spec.spec_id
        return new_spec

    def _load(self, **local_inputs):
        """Transfer class variables to instance."""
        for key in self.__class__.__dict__:
            if (
                key.startswith("__")
                or key == "_INTERNAL_FIELDS"
                or key in self._INTERNAL_FIELDS
            ):
                continue

            spec = getattr(self.__class__, key)

            # Skip partial kwargs (no registration needed)
            if isinstance(spec, dict):
                continue

            if not isinstance(spec, specs.Spec):
                raise ValueError(
                    f"Expected Spec type, got {type(spec)} with {key!r}"
                )

            # Register key
            self._keys[spec.spec_id] = key

            # Handle inputs
            if isinstance(spec, specs.GlobalInput):
                self._global_inputs[key] = spec.spec_id

                spec = self._process_input(
                    key, spec, self._config_locator.global_inputs, "Global"
                )
            elif isinstance(spec, specs.LocalInput):
                spec = self._process_input(key, spec, local_inputs, "Local")

            # Handle child configs
            if isinstance(spec, ConfigSpec):
                child_config = self._config_locator.get(spec)
                self._child_configs[key] = child_config
            else:
                self._specs[key] = spec

        self._loaded = True

    def freeze(self):
        """Freeze to prevent any more perturbations to this Config instance."""
        self._frozen = True

    # NB: Have to override getattribute instead of getattr so as to
    # prevent initial, class-level values from being used.
    def __getattribute__(self, key: str) -> Union[specs.Spec, Config]:
        if (
            key.startswith("__")
            or key == "_INTERNAL_FIELDS"
            or key in self._INTERNAL_FIELDS
        ):
            return super().__getattribute__(key)

        try:
            if key in self._child_configs:
                return self._child_configs[key]
            else:
                return self._specs[key]
        except KeyError:
            raise KeyError(f"{self.__class__}: {key!r}")

    def __getitem__(self, key: str) -> Any:
        return utils.nested_getattr(self, key)

    def __setattr__(self, key: str, value: Any):
        if (
            key.startswith("__")
            or key == "_INTERNAL_FIELDS"
            or key in self._INTERNAL_FIELDS
        ):
            super().__setattr__(key, value)
            return

        if self._frozen:
            raise errors.FrozenConfigError(
                f"Cannot perturb frozen config: key={key!r}"
            )

        if key not in self._specs and self._loaded:
            if key in self._child_configs:
                raise errors.SetChildConfigError(
                    f"Cannot set child config: key={key!r}"
                )
            else:
                raise errors.NewKeyConfigError(
                    f"Cannot add new keys to a loaded config: key={key!r}"
                )

        old_spec = self._specs[key]

        if not isinstance(value, specs.Spec):
            value = specs.Object(value)

        self._specs[key] = value
        value.spec_id = old_spec.spec_id

    def __setitem__(self, key: str, value: Any):
        utils.nested_setattr(self, key, value)

    def __dir__(self) -> Iterable[str]:
        return sorted(
            list(self._specs.keys()) + list(self._child_configs.keys())
        )


class ConfigLocator:
    """Service locator to get instances of Configs by types."""

    def __init__(self, **global_inputs):
        self.global_inputs: Dict[str, Any] = global_inputs

        self._config_cache: Dict[ConfigSpec, Config] = {}

    def get(self, config_spec: ConfigSpec) -> Config:
        """Get Config instance by type."""
        try:
            return self._config_cache[config_spec]
        except KeyError:
            pass

        # noinspection PyProtectedMember
        config = config_spec._instantiate(self)
        self._config_cache[config_spec] = config
        return config



from __future__ import annotations

from typing import Any, Callable, Dict, List, Optional, Type, Union

MISSING = object()
MISSING_DICT: Dict = {}
SpecID = int


def instantiate(cls: type, *args, **kwargs) -> Any:
    """Instantiate obj from Spec parts."""
    obj: Any = object.__new__(cls)
    try:
        obj.__init__(*args, **kwargs)
    except TypeError as exc:
        raise TypeError(f"{cls}: {str(exc)}")
    return obj


class AttrFuture:
    """Future representing attr access on a Spec."""

    def __init__(self, parent_spec_id: SpecID, attrs: List[str]):
        self.parent_spec_id = parent_spec_id
        self.attrs = attrs

    def __getattr__(self, attr: str) -> AttrFuture:
        return AttrFuture(self.parent_spec_id, self.attrs + [attr])


class Spec:
    """Represents delayed obj to be instantiated later."""

    NEXT_SPEC_ID = 0

    def __init__(self):
        self.spec_id = self._get_next_spec_id()

    def __getattr__(self, attr: str) -> AttrFuture:
        return AttrFuture(self.spec_id, [attr])

    # For mypy
    def __call__(self, *args, **kwargs):
        return None

    @staticmethod
    def _get_next_spec_id() -> SpecID:
        result = Spec.NEXT_SPEC_ID
        Spec.NEXT_SPEC_ID += 1
        return result


class Object(Spec):
    """Represents fully-instantiated obj."""

    def __init__(self, obj: Any):
        super().__init__()
        self.obj = obj


class Input(Spec):
    """Represents global input that can be set while getting Config."""

    def __init__(self, type_: Optional[Type] = None, default: Any = MISSING):
        super().__init__()
        self.type_ = type_
        self.default = default


class GlobalInput(Input):
    pass


class LocalInput(Input):
    pass


class Prototype(Spec):
    """Represents obj to be instantiated at every Container get call."""

    def __init__(self, cls: Union[Type, Callable], *args, **kwargs):
        super().__init__()
        self.cls = cls
        self.args = args
        self.lazy_kwargs = kwargs.pop("__lazy_kwargs", None)
        self.kwargs = kwargs

    def instantiate(self) -> Any:
        """Instantiate, useful when using outside DI framework."""
        if isinstance(self.cls, type):
            return instantiate(self.cls, *self.args, **self.kwargs)
        else:
            # Non-type callable (e.g., function, functor)
            return self.cls(*self.args, **self.kwargs)

    def copy_with(self, *args, **kwargs) -> Prototype:
        return self.__class__(self.cls, *args, **kwargs)


class PrototypeIdentity(Prototype):
    """Represents forwarding to another spec."""

    def __init__(self, obj: Any):
        super().__init__(lambda obj_: obj_, obj)

    def copy_with(self, *args, **kwargs) -> PrototypeIdentity:
        return self.__class__(*args, **kwargs)


Forward = PrototypeIdentity


class Singleton(Prototype):
    """Represents obj to be instantiated once per key per Container."""

    pass


class SingletonCollection(Singleton):
    """Represents collection to be instantiated once per key per Container."""

    def copy_with(self, *args, **kwargs) -> SingletonCollection:
        return self.__class__(*args, **kwargs)


class SingletonTuple(SingletonCollection):
    """Represents tuple to be instantiated once per key per Container."""

    def __init__(self, *args):
        super().__init__(lambda *args_: args_, *args)


class SingletonList(SingletonCollection):
    """Represents list to be instantiated once per key per Container."""

    def __init__(self, *args):
        super().__init__(lambda *args_: list(args_), *args)


class SingletonDict(SingletonCollection):
    """Represents dict to be instantiated once per key per Container.

    Can be used either with kwargs:

        >>> values = dilib.SingletonDict(x=x, y=y)

    Or with a dict (useful when keys are not strs):

        >>> values = dilib.SingletonDict({1: x, 2: y})
    """

    def __init__(self, values: Dict = MISSING_DICT, **kwargs):
        if values is MISSING_DICT:
            super().__init__(lambda **kwargs_: kwargs_, **kwargs)
            self._kwargs_style = True
        else:
            if kwargs:
                raise ValueError("Cannot set both values and kwargs")

            super().__init__(
                lambda values_: values_, __lazy_kwargs=dict(values_=values)
            )
            self._kwargs_style = False

    def copy_with(self, *args, **kwargs) -> SingletonCollection:
        if self._kwargs_style:
            return super().copy_with(*args, **kwargs)
        else:
            return Singleton(lambda values_: values_["values_"], kwargs)  # type: ignore  # noqa


class PrototypeMixin:
    """Helper class for Prototype to ease syntax in Config."""

    def __new__(cls: type, *args, **kwargs):
        return Prototype(cls, *args, **kwargs)


class SingletonMixin:
    """Helper class for Singleton to ease syntax in Config."""

    def __new__(cls: type, *args, **kwargs):
        return Singleton(cls, *args, **kwargs)
